/*
 GBxCart RW - Console Interface Flasher
 Version: 1.16
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 26/08/2017
 Last Modified: 2/03/2019
 
 This program allows you to write ROM to Flash Carts that are supported.
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

int main(int argc, char **argv) {
	
	printf("GBxCart RW Flasher v1.16 by insideGadgets\n");
	printf("########################################\n");
	
	// Check arguments
	if (argc >= 2) {
		read_config();
		
		// Open COM port
		if (com_test_port() == 0) {
			printf("Device not connected and couldn't be auto detected\n");
			read_one_letter();
			return 1;
		}
		printf("Connected on COM port: %i\n", cport_nr+1);
		
		// Break out of any existing functions on ATmega
		set_mode('0');
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		// Get firmware version
		gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
		
		// Get PCB version
		gbxcartPcbVersion = request_value(READ_PCB_VERSION);
		
		if (gbxcartPcbVersion == PCB_1_0) {
			printf("\nPCB v1.0 is not supported for this function.");
			read_one_letter();
			return 1;
		}
		
		if (gbxcartFirmwareVersion <= 8) {
			printf("Firmware R9 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		
		// Separate the file name
		char filenameOnly[100];
		char filename[254];
		strncpy(filename, argv[1], 253);
		
		char *token = strtok(filename, "\\");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "\\");
		}
		
		printf("\n--- Write ROM to Flash Cart ---\n");
		printf("Cart selected: ");
		
		uint32_t readBytes = 0;
		uint16_t sector = 0;
		long fileSize = 0;
		
		// Grab file size
		FILE *romFile = fopen(argv[1], "rb");
		if (romFile != NULL) {
			fseek(romFile, 0, SEEK_END);
			fileSize = ftell(romFile);
			fseek(romFile, 0, SEEK_SET);
		}
		else {
			printf("\n%s \nFile not found\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		printf("16 MByte M36L0R706 / 32 MByte 256L30B Gameboy Advance Flash Cart\n");
		
		
		printf("\nGoing to write to ROM (Flash cart) from %s\n", filenameOnly);
		
		// Check file size
		if (fileSize > 0x2000000) {
			fclose(romFile);
			printf("\n%s \nFile size is larger than the available Flash cart space of 32 MBytes\n", argv[1]);
			read_one_letter();
			return 1;
		}
		
		printf("\nWriting to ROM (Flash cart) from %s\n", filenameOnly);
		printf("[             25%%             50%%             75%%            100%%]\n[");
		
		
		// Set to reading mode
		currAddr = 0x0000;
		gba_flash_write_address_byte(currAddr, 0xFF);
		delay_ms(5);
		
		// Read rom a tiny bit before writing
		currAddr = 0x0000;	
		set_number(currAddr, SET_START_ADDRESS);
		set_mode(GBA_READ_ROM);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop();
		
		// Set end address as file size
		endAddr = fileSize;
		
		
		// Flash ID command
		gba_flash_write_address_byte(0x00, 0x90);
		delay_ms(1);
		
		// Read ID
		set_number(0x00, SET_START_ADDRESS);
		set_mode(GBA_READ_ROM);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop(); // End read
		
		// For chips with 4x 16K sectors at the start
		// 256L30B (0x8A, 0x0, 0x15, 0x88)
		int sectorEraseAddress = 0x20000;
		if (readBuffer[0] == 0x8A && readBuffer[1] == 0 && readBuffer[2] == 0x15 && readBuffer[3] == 0x88) {
			sectorEraseAddress = 0x8000;
		}
		
		// Back to reading mode
		gba_flash_write_address_byte(currAddr, 0xFF);
		delay_ms(5);
		
		
		// Write ROM
		currAddr = 0x0000;	
		set_number(currAddr, SET_START_ADDRESS);
		uint32_t addressForManualWrite = 0x7FC0;
		
		while (currAddr < endAddr) {
			if (currAddr % sectorEraseAddress == 0) { // Erase next sector
				// For chips with 4x 16K sectors, after 64K, change to 0x20000 sector size
				if (currAddr >= 0x20000 && sectorEraseAddress == 0x8000) {
					sectorEraseAddress = 0x20000;
				}
				
				//printf("Erasing sector at 0x%X\n", currAddr);
				
				// Unlock
				gba_flash_write_address_byte(currAddr, 0x60);
				gba_flash_write_address_byte(currAddr, 0xD0);
				
				// Erase
				gba_flash_write_address_byte(currAddr, 0x20);
				gba_flash_write_address_byte(currAddr, 0xD0);
				delay_ms(50);
				
				// Wait for first 2 bytes to be 0x80, 0x00
				readBuffer[0] = 0;
				readBuffer[1] = 0;
				while (readBuffer[0] != 0x80 || readBuffer[1] != 0x00) {
					set_number(currAddr / 2, SET_START_ADDRESS);
					delay_ms(5);
					set_mode(GBA_READ_ROM);
					delay_ms(5);
					
					com_read_bytes(READ_BUFFER, 64);
					com_read_stop(); // End read
					
					//printf("0x%X,0x%X ", readBuffer[0], readBuffer[1]);
					delay_ms(500);
				}
				
				// Back to reading mode
				gba_flash_write_address_byte(currAddr, 0xFF);
				delay_ms(5);
				
				set_number(currAddr / 2, SET_START_ADDRESS); // Divide address by 2
				delay_ms(5);
			}
			
			// Standard buffered writing
			
				/*com_write_bytes_from_file(GBA_FLASH_WRITE_INTEL_64BYTE, romFile, 64);
				com_wait_for_ack();
				currAddr += 64;
				readBytes += 64;
			*/
			
			// It seems that some carts don't allow us to use the buffered 32 byte programming mode for 0x7FC0 to 0x7FFF and 
			// it also happens at 0xFFC0 to 0xFFFF and so on. We just write each of the 32 bytes one byte at a time.
			if (currAddr > 0 && currAddr == addressForManualWrite) {
				//printf("Address Manual writing 0x%X\n", currAddr);
				//printf(".");
				uint8_t localbuffer[64];
				fread(&localbuffer, 1, 64, romFile);
				
				for (uint8_t x = 0; x < 64; x += 2) {
					uint16_t combinedBytes = (uint16_t) localbuffer[x+1] << 8 | (uint16_t) localbuffer[x];
					gba_flash_write_address_byte(currAddr, 0x40);
					delay_ms(1);
					gba_flash_write_address_byte(currAddr, combinedBytes);
					delay_ms(1);
					currAddr += 2;
					readBytes += 2;
				}
				addressForManualWrite += 0x8000;
				
				// Set address again (seems to be needed at and after 0x27C0)
				set_number(currAddr / 2, SET_START_ADDRESS); // Divide address by 2
				delay_ms(5);
			}
			else {
				//printf("Address 0x%X\n", currAddr);
				com_write_bytes_from_file(GBA_FLASH_WRITE_INTEL_64BYTE, romFile, 64);
				com_wait_for_ack();
				currAddr += 64;
				readBytes += 64;
			}
			
			print_progress_percent(readBytes, (endAddr) / 64);
		}
		
		// Back to reading mode
		gba_flash_write_address_byte(currAddr, 0xFF);
		delay_ms(5);
		
	}
	
	printf("\n");
	
	return 0;
}